/** 
 *   DGPS - Data.
 *
 *   (c) Copyright 2007 Federico Bareilles <fede@iar.unlp.edu.ar>,
 *   Instituto Argentino de Radio Astronomia (IAR).
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 2 of
 *   the License, or (at your option) any later version.
 *     
 *   The author does NOT admit liability nor provide warranty for any
 *   of this software. This material is provided "AS-IS" in the hope
 *   that it may be useful for others.
 *
 **/

#define _GNU_SOURCE

#include <errno.h> 
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <limits.h>
#include <syslog.h>
#include <ctype.h>

#include <my_stdio.h>

#include "gps_rs232.h"
#include "dgps_conf.h"
#include "dgps_data.h"


/**
ASCII LOG STRUCTURE 

 Log types ending with the letter A (or a) will be output in ASCII
 format (e.g., POSA). The structures of all ASCII logs follow the
 general conventions as noted here:

   1. The lead code identifier for each record is '$'.

   2. Each log is of variable length depending on amount of data and formats.

   3. All data fields are delimited by a comma ',' with the exception
   of the last data field, which is followed by a * to indicate end of
   message data.

   4. Each log ends with a hexadecimal number preceded by an asterisk
   and followed by a line termination using the carriage return and
   line feed characters, e.g., *xx[CR][LF]. This 8-bit value is an
   exclusive OR (XOR) of all bytes in the log, excluding the '$'
   identifier and the asterisk preceding the two checksum digits.

 Structure: 
   $xxxx, data field..., data field..., data field... *xx [CR][LF]
 **/

/**

help
?              -Online Command Help    ACCEPT         -Accept Datatypes
ANTENNAPOWER   -Antenna Power Control  ASSIGN         -Assign PRN To a Chan.
CLOCKADJUST    -Adjust 1pps            COM1           -Intialize Port 1
COM2           -Intialize Port 2       COM1_DTR       -DTR Control on Port 1
COM2_DTR       -DTR Control on Port 2  COM1_RTS       -RTS Control on Port 1
COM2_RTS       -RTS Control on Port 2  CONFIG         -Configure Satellites
CRESET         -Factory Config Reset   CSMOOTH        -Carrier Smoothing
DATUM          -Choose a DATUM Type    DGPSTIMEOUT    -Max age of DGPS data
DIFF_PROTOCOL  -Diff. protocol control DYNAMICS       -Set Receiver Dynamics
ECUTOFF        -Elevation Cutoff Angle EXTERNALCLOCK  -Specify Clock type
FIX            -Set Antenna Coord.     FRESET         -Factory Card Reset
FREQUENCY_OUT  -Variable Freq. Output  HELP           -Online Command Help
LOCKOUT        -Lock Out Satellite     LOG            -Choose Data Logging
MAGVAR         -Set Magnetic Variation MESSAGES       -Error Messages ON/OFF
POSAVE         -Position Averaging     RESET          -Hardware Reset
RESETHEALTH    -Reset PRN Health       RESETHEALTHALL -Reset All PRN Health
RESETRT20      -Reset rt20 algorithm   RTKMODE        -Set RTK parameters
RINEX          -RINEX Configuration    RTCM16T        -Input Type 16 Message
RTCMRULE       -Rtcm Bit Rule          SAVEALMA       -Save Almanac & ION/UTC
SAVECONFIG     -Save User Config.      SEND           -Send string to a port
SENDHEX        -Send hex to a port     SETDGPSID      -Set the Station ID
SETHEALTH      -Override PRN Health    SETL1OFFSET    -Set L1 PSR Offset
SETNAV         -Set a Destination      SETTIMESYNC    -Enable/Disable Timesync
UNASSIGN       -Un-Assign a Channel    UNASSIGNALL    -Un-Assign All Channels
UNDULATION     -Choose Undulation      UNFIX          -Remove Recvr. FIX(ed)
UNLOCKOUT      -Restore Satellite      UNLOCKOUTALL   -Select All Satellites
UNLOG          -Kill a Data Log        UNLOGALL       -Kill All Data Logs
USERDATUM      -User Defined DATUM     VERSION        -Current Software Ver.
Com1>


VERSION
OEM-3 MILLENSTD CGL97110081 HW 3-1 SW 4.433/2.03 Feb 18/97
Com1>

**/

/**
FRESET 

  This command clears all data that is stored in non-volatile
  memory. Such data includes the almanac, satellite channel
  configuration, and any user-specific configurations. It is only
  available on 12-channel OEM cards with the SAVECONFIG option. The
  GPSCard is forced to reset and will start up with factory
  defaults. See also the commands CRESET, page 25, and RESET, page
  40. 

Syntax: 
  FRESET

SAVECONFIG
 
  This command saves the current configuration and up to 10 active
  logging commands in flash (non-volatile) memory. It is only
  available on 12-channel OEM cards with the SAVECONFIG option. The
  last saved configuration becomes the new power-on default overriding
  the factory default. Between 9 and 18 SAVECONFIG commands can be
  issued before a reset is necessary. If a SAVECONFIG is not possible,
  the error message Flash full RESET or use FORCE option will be
  given. If at this point the user enters SAVECONFIG FORCE , the
  configuration will be saved and the card will be reset. If the flash
  memory is not full, the FORCE option has no effect. SAVECONFIG FORCE
  will simply cause a normal save and no reset will occur. A CRESET
  command may be entered to restore the factory default settings,
  which will become the new saved configuration if the user enters
  another SAVECONFIG command. The RCCA log can be used to monitor the
  current configuration. 

Syntax: 
  SAVECONFIG 
or 
  SAVECONFIG FORCE


RINEX OPTION <value>

Configures RINEX log user defined fields

OPTION   = COMMENT        - Add comment to navigation & observation log headers
         = AGENCY         - Define name of agency in observation log header
         = MARKNAME       - Define marker name in observation log
         = MARKERNUM      - Define marker number in observation log
         = OBSERVER       - Define name of observer in observation log header
         = RECNUM         - Define receiver number in observation log header
         = ANTNUM         - Define antenna number in observation log header
         = ANTTYPE        - Define antenna type in observation log header
         = ANTDN          - Define antenna delta North in observation log
         = ANTDE          - Define antenna delta East in observation log
         = ANTDH          - Define antenna delta Height in observation log

rinex OBSERVER Mauricio Gende
log com1 xohd


log posa
Com1>
$POSA,428,519944.50,-34.90742829,-57.93305075,6.462,17.948,61,64.782,9.954,10.40

log posa
Com1>
$POSA,0,0.00,0.00000000,0.00000000,-6378053.700,16.700,61,0.000,0.000,0.000,1*01


**/


static int wait_char(int fd, char patern)
{
	int ret = 0;
	char c;
	
	do {
		ret = hw_read(fd, &c, 1);
	} while ((ret == 1) && c != patern);
	if (ret == 1) ret = 0;

	return ret;
}


static int wait_prompt(int fd)
{
	return wait_char(fd, '>');
}


int dgps_get_line(int fd, struct data_line *dl)
{
	int i;
	char *line = dl->line;
	int ret = 0;
	int end = 0;
	unsigned char chk = ',';
	int len_code = 0;
	int pk_chk = 0;
	
	dl->len = 0;
	dl->code = LEAD_CODE_UNKNOW;
	dl->valid_flag = 0;

	if (wait_char(fd, '$') != 0) {
		PVERB(PWARN, "whiting for '$' time out");
		return -1;
	}
	i = 0;
	do {
		ret = hw_read(fd, line+i, 1);
		PVERB(PINFO,"%c", *(line+i));
		i++;
	} while ((ret == 1) && *(line+i-1) != ',' && (i < MAX_BUFF_LEN));
	if (ret != 1) {
		PVERB(PWARN, "whiting for ',' time out");
		return -1;
	}
	/* Aca ya esta el lead code obtenido: */
	len_code = i - 1; /* pero debe ser sienpre 4 */
	if ( strncmp(line, "XOBS", 4) == 0) {
		dl->code = LEAD_CODE_XOBS;
	} else if ( strncmp(line, "XOHD", 4) == 0) {
		dl->code = LEAD_CODE_XOHD;
	} /* Add more lead code hear */

	for (i=0;i<len_code;i++) 
		chk ^= (unsigned char) *(line+i);
	dl->len = 0;
	do {
		ret = hw_read(fd, line+dl->len, 1);
		if (ret == 1) dl->len+= ret;
	} while ((ret == 1) && *(line+dl->len-1) != '*');
	if ( ret != 1) {
		PVERB(PWARN, "whiting for '*' time out");
		return -1;
	}	
	end = dl->len - 1;
	ret = hw_read(fd, line+dl->len, 2);
	if (ret == 2){
		dl->len+= ret;
		*(line+dl->len) = 0;
		for (i=0;i<end;i++) 
			chk ^= (unsigned char) *(line+i);
		pk_chk = (int) strtoul(line+end+1, NULL, 16);
		if ( pk_chk == chk) dl->valid_flag = 1;
		/* descartad 0x0d 0x0a: */
		//ret = hw_read(fd, &c, 1);
		//ret = hw_read(fd, &c, 1);
	}
	if ( ret > 0) ret = 0;
	line[end] = 0;
	dl->len = end;

	return ret;
}


struct rinex_heder *rinex_heder_get_conf(char *file)
{
	struct rinex_heder *rh= NULL;
	struct configuration *conf = NULL;
        struct configuration *tag = NULL;

	if (read_conf_file(&conf, file) != 0) {
		return NULL;
	}
	rh = (struct rinex_heder *) malloc(sizeof(struct rinex_heder));
	rh->comment = NULL;   
	rh->agency = NULL;	 
	rh->markername = NULL;   
	rh->markernum = NULL;    
	rh->observer = NULL;	 
	rh->recnum = NULL;	 
	rh->antnum = NULL;    
	rh->antype = NULL;	 
	rh->antdn = NULL;	     
	rh->antde = NULL;	 
	rh->antdh = NULL;	 
	rh->conf = conf;
	if ((tag = find_conf_tag(conf, "COMMENT")) != NULL) {
		rh->comment = tag;
	}
	if ((tag = find_conf_tag(conf, "AGENCY")) != NULL) {
		rh->agency = tag;
	}
	if ((tag = find_conf_tag(conf, "MARKNAME")) != NULL) {
		rh->markername = tag;
	}
	if ((tag = find_conf_tag(conf, "MARKERNUM")) != NULL) {
		rh->markernum = tag;
	}
	if ((tag = find_conf_tag(conf, "OBSERVER")) != NULL) {
		rh->observer = tag;
	}
	if ((tag = find_conf_tag(conf, "RECNUM")) != NULL) {
		rh->recnum = tag;
	}
	if ((tag = find_conf_tag(conf, "ANTNUM")) != NULL) {
		rh->antnum = tag;
	}
	if ((tag = find_conf_tag(conf, "ANTTYPE")) != NULL) {
		rh->antype = tag;
	}
	if ((tag = find_conf_tag(conf, "ANTDN")) != NULL) {
		rh->antdn = tag;
	}
	if ((tag = find_conf_tag(conf, "ANTDE")) != NULL) {
		rh->antde = tag;
	}
	if ((tag = find_conf_tag(conf, "ANTDH")) != NULL) {
		rh->antdh = tag;
	}

	return rh;
}

	
int free_rinex_heder(struct rinex_heder *rh)
{
	if (rh->conf != NULL) free_conf(rh->conf);
	free(rh);

	return 0;
}


static int set_rinex_heder_from_tag(int fd, struct configuration *tag)
{
	if (tag != NULL)
		if (strlen(tag->data) > 0)
			fd_printf(fd, "rinex %s %s\n\r", tag->tag, tag->data);
	
	return 0;
}


int set_rinex_heder(int fd, struct rinex_heder *rh)
{
	if (rh->comment != NULL)
		set_rinex_heder_from_tag(fd, rh->comment);
	if (rh->agency != NULL)	 
		set_rinex_heder_from_tag(fd, rh->agency);
	if (rh->markername != NULL)  
		set_rinex_heder_from_tag(fd, rh->markername);
	if (rh->markernum != NULL)    
		set_rinex_heder_from_tag(fd, rh->markernum);
	if (rh->observer != NULL)	 
		set_rinex_heder_from_tag(fd, rh->observer);
	if (rh->recnum != NULL)	 
		set_rinex_heder_from_tag(fd, rh->recnum);
	if (rh->antnum != NULL)    
		set_rinex_heder_from_tag(fd, rh->antnum);
	if (rh->antype != NULL)	 
		set_rinex_heder_from_tag(fd, rh->antype);
	if (rh->antdn != NULL)	     
		set_rinex_heder_from_tag(fd, rh->antdn);
	if (rh->antde != NULL)	 
		set_rinex_heder_from_tag(fd, rh->antde);
	if (rh->antdh != NULL)	 
		set_rinex_heder_from_tag(fd, rh->antdh);

	return 0;
}


int save_flash(int fd)
{
	SYSLOG(LOG_INFO, "Save current configuration in non-volatile memory");
	fd_printf(fd, "SAVECONFIG FORCE\n\r");

	return wait_prompt(fd);
}


int reset_conf(int fd)
{
	SYSLOG(LOG_INFO,
	       "Reseting non-volatile memory (almanac, satellite channel \
configuration, and any user-specific configurations)")
	fd_printf(fd, "FRESET\n\r");

	return wait_prompt(fd);
}


int set_logging(int fd, struct dgps_conf *dc)
{
	SYSLOG(LOG_INFO, "Setting loggin mode");
	fd_printf(fd, "log %s xobs ontime 1\n\r", dc->log_port);

	return wait_prompt(fd);
}


int set_unlogging(int fd)
{
	SYSLOG(LOG_INFO, "Unlog all");
	fd_printf(fd, "UNLOGALL\n\r");

	return wait_prompt(fd);


}


int line_is_epoch( char *line)
{
	int ret = 0;

	if(isdigit( (int) *(line)) && (*(line+32) == 'G'))
	   ret = 1;

	return ret;
}

int parse_epoch_line( char *line, struct epoch *ep)
{
	int i, j;

/* line type:
|-------------------------------|
00000000001111111111222222222233333333334444444444555555555566666666667777777777
01234567890123456789012345678901234567890123456789012345678901234567890123456789
107 11 01 03 06 56.0000000  0  6G 5G 9G12G17G26G 4

*/
	if(sscanf(line,"%3d %2d %2d %2d %2d%11lf  %1d%3d",
		  &ep->year, 
		  &ep->mon , &ep->mday, 
		  &ep->hour, &ep->min, &ep->sec,
		  &ep->ep,
		  &ep->num_sat) == 8) {
		ep->year = ep->year - 100;
		for(i=0, j = 32; i < ep->num_sat; i++, j+=3) {
			sscanf(line+j, "G%2d", &ep->sat_list[i]);
			ep->sat_valid_list[i] = 0;
		}

		return 0;
	} else {
		PVERB(PWARN,"no line epoch.");
	}
	
	return -1;
}


int print_epoch(int fd, struct epoch *ep)
{
	int i;
	int valid = 0;


	for(i=0; i < ep->num_sat; i++)
		if(ep->sat_valid_list[i] == 2) valid ++;
	if (valid) {
		fd_printf(fd, " %02d %02d %02d %02d %02d%11.7f  %1d%3d",
			  ep->year, 
			  ep->mon , ep->mday, 
			  ep->hour, ep->min, ep->sec,
			  ep->ep, valid /*ep->num_sat*/);
		for(i=0;i<ep->num_sat;i++)
			if(ep->sat_valid_list[i] == 2) 
				fd_printf(fd, "G%2d", ep->sat_list[i]);
		fd_printf(fd, "\n");
	
		/* FIXME: imprimir los daos de los satelites */
		for(i=0;i<ep->num_sat;i++)
			if(ep->sat_valid_list[i] == 2) {
				fd_printf(fd, "%s\n", ep->sat[(i<<1)]);
				fd_printf(fd, "%s\n", ep->sat[(i<<1)|1]);
			}
	}

	return 0;
}
